use std::io::Cursor;

use anyhow::anyhow;
use log::{info, debug, warn};
use serde_json::json;
use tokio::{net::{TcpStream, tcp::{OwnedReadHalf, OwnedWriteHalf}}, select, io::{AsyncReadExt, AsyncWriteExt}};
use uuid::Uuid;

use crate::{varint::VarInt, ser::{Deserializable, Position}, protocol::{self, Protocol, ProtocolState, ClientPacket, ServerPacket}, Player, ClientInfo, JsonValue, HTTP_CLIENT};
use event::{ClientEvent, ServerEvent};

pub mod event;

type Sender = tokio::sync::mpsc::Sender<(i32, ClientEvent)>;
type Receiver = tokio::sync::mpsc::Receiver<ServerEvent>;

// arbitrary
const OFFLINE_NAMESPACE: Uuid = Uuid::from_bytes([0xc3, 0xe3, 0xe7, 0xa5, 0x58, 0xe5, 0x4c, 0xf4, 0x84, 0xef, 0x27, 0x4b, 0x9e, 0xf1, 0x5c, 0xc3]);

async fn fetch_uuid(name: &str) -> anyhow::Result<Uuid> {
	let url = "https://api.mojang.com/users/profiles/minecraft/".to_owned() + name;
	let response = HTTP_CLIENT.get().unwrap()
		.get(url).send().await?;
	let status = response.status();
	let json: JsonValue = response.json().await?;
	if !status.is_success() || status.as_u16() == 204 {
		return Err(match json["errorMessage"].as_str() {
			Some(e) => anyhow!("request failed: {}", e),
			None => anyhow!("unknown error in request"),
		})
	}
	let uuid = json["id"].as_str().ok_or_else(|| anyhow!("response did not contain id field"))?;
	Ok(Uuid::parse_str(uuid)?)
}

async fn get_offline_uuid(name: &str) -> Uuid {
	match fetch_uuid(name).await {
		Ok(uuid) => {
			debug!("fetched uuid for {name}: {uuid}");
			uuid
		}
		Err(e) => {
			warn!("error fetching uuid for {name}: {e}");
			Uuid::new_v5(&OFFLINE_NAMESPACE, name.as_bytes())
		}
	}
}

async fn read_varint_async(r: &mut OwnedReadHalf) -> std::io::Result<Option<i32>> {
	let mut buf = [0];
	let mut result = 0u32;
	for count in 0..5 {
		if r.read(&mut buf).await? != 1 {
			return Ok(None)
		}
		let byte = buf[0];
		result |= ((byte & 0x7f) as u32) << (7 * count);
		if byte & 0x80 == 0 {
			return Ok(Some(result as i32))
		}
	}
	Ok(None)
}

async fn write_varint_async(w: &mut OwnedWriteHalf, i: i32) -> std::io::Result<()> {
	let mut data = i as u32;
	loop {
		let mut byte = (data & 0x7F) as u8;
		data >>= 7;
		if data != 0 {
			byte |= 0x80;
		}
		w.write_u8(byte).await?;
		if data == 0 {
			return Ok(())
		}
	}
}

struct ClientState {
	id: i32,
	proto: Protocol,
	state: ProtocolState,
	r: OwnedReadHalf,
	w: OwnedWriteHalf,
	buf: Vec<u8>
}

enum ReadPacketOutcome {
	Packet(ClientPacket),
	None,
	Eof,
}

pub async fn run_client(id: i32, stream: TcpStream, tx: Sender, rx: Receiver) -> anyhow::Result<()> {
	debug!("running client #{id}");
	let (r, w) = stream.into_split();
	let client = ClientState {
		id,
		proto: protocol::COMMON,
		r, w,
		buf: Vec::new(),
		state: ProtocolState::Handshake,
	};
	client.handshake(tx, rx).await
}

impl ClientState {
	async fn read_packet(&mut self) -> anyhow::Result<ReadPacketOutcome> {
		match read_varint_async(&mut self.r).await? {
			Some(len) => {
				if len as usize > self.buf.len() {
					self.buf.resize(len as usize, 0);
				}
				self.r.read_exact(&mut self.buf[0..(len as usize)]).await?;
				let mut cursor = Cursor::new(&mut self.buf);
				let id = VarInt::deserialize(&mut cursor)?.0;
				let decoded = (self.proto.decode)(Box::new(&mut cursor), self.state, len, id)?;
				Ok(match decoded {
					Some(pk) => ReadPacketOutcome::Packet(pk),
					None => ReadPacketOutcome::None,
				})
			}
			None => Ok(ReadPacketOutcome::Eof),
		}
	}

	async fn write_packet(&mut self, pk: ServerPacket) -> anyhow::Result<()> {
		self.buf.clear();
		let mut cursor = Cursor::new(&mut self.buf);
		(self.proto.encode)(Box::new(&mut cursor), self.state, pk)?;
		write_varint_async(&mut self.w, self.buf.len() as i32).await?;
		self.w.write_all(&self.buf).await?;
		Ok(())
	}


	async fn handshake(mut self, tx: Sender, rx: Receiver) -> anyhow::Result<()> {
		let ReadPacketOutcome::Packet(ClientPacket::Handshake { version, addr, port, next_state }) = self.read_packet().await? else {
			return Err(anyhow!("Did not recieve handshake"))
		};

		self.state = ProtocolState::try_from(next_state).unwrap_or(ProtocolState::Handshake);

		if let Some(proto) = protocol::get_protocol(version) {
			tx.send((self.id, ClientEvent::Handshake(ClientInfo {
				addr,
				port,
				proto_version: proto.version,
				proto_name: proto.name,
			}))).await?;
			self.proto = proto;

			info!("#{} using protocol {} ({})", self.id, proto.name, proto.version);
		} else {
			info!("#{} using unsupported protocol {}", self.id, version);
		}

		match self.state {
			ProtocolState::Status => self.slp(tx, rx).await,
			ProtocolState::Login => self.login(tx, rx).await,
			ProtocolState::Handshake
			| ProtocolState::Play => Err(anyhow!("invalid next state")),
		}
	}

	async fn slp(mut self, _tx: Sender, mut rx: Receiver) -> anyhow::Result<()> {
		debug!("#{} entering slp", self.id);
		loop { select! {
			ev = rx.recv() => {
				if let Some(ServerEvent::Disconnect(msg)) = ev {
					debug!("#{} disconnecting: {}", self.id, msg.unwrap_or_default());
					return Ok(())
				}
			},
			pk = self.read_packet() => match pk? {
				ReadPacketOutcome::Packet(ClientPacket::PingRequest(data)) => {
					debug!("#{} ping request: {}", self.id, data);
					self.write_packet(ServerPacket::PingResponse(data)).await?;
				},
				ReadPacketOutcome::Packet(ClientPacket::StatusRequest) => {
					debug!("#{} status request", self.id);
					let desc = if self.proto == protocol::COMMON {
						"Unsupported client version"
					} else {
						"server list ping"
					};
					self.write_packet(ServerPacket::StatusResponse(json!(
						{
							"version": {
								"name": self.proto.name,
								"protocol": self.proto.version,
							},
							"players": {
								"max": 1337,
								"online": 0,
							},
							"description": {
								"text": desc,
							},
							"enforcesSecureChat": false,
							"previewsChat": false,
						}
					))).await?;
				},
				ReadPacketOutcome::Eof => {
					debug!("#{} got eof, closing", self.id);
					return Ok(())
				},
				_ => (),
			},
		} }
	}

	async fn login(mut self, tx: Sender, mut rx: Receiver) -> anyhow::Result<()> {
		debug!("#{} entering login", self.id);
		if self.proto == protocol::COMMON {
			self.write_packet(ServerPacket::LoginDisconnect(json!(
				{
					"text": "Unsupported client version"
				}
			))).await?;
			debug!("#{} disconnecting due to unsupported version", self.id);
			return Ok(())
		}
		loop { select! {
			ev = rx.recv() => {
				if let Some(ServerEvent::Disconnect(msg)) = ev {
					self.write_packet(ServerPacket::LoginDisconnect(json!(
						{
							"text": msg
						}
					))).await?;
					info!("#{} disconnecting: {}", self.id, msg.unwrap_or_default());
					return Ok(())
				}
			},
			pk = self.read_packet() => match pk? {
				ReadPacketOutcome::Packet(ClientPacket::LoginStart { name, uuid }) => {
					let uuid = match uuid {
						Some(uuid) => uuid,
						None => {
							debug!("#{} no uuid provided, fetching", self.id);
							get_offline_uuid(&name).await
						}
					};
					info!("#{} logged in as {name} ({uuid})", self.id);
					self.write_packet(ServerPacket::LoginSuccess { name: name.clone(), uuid }).await?;
					self.state = ProtocolState::Play;
					tx.send((self.id, ClientEvent::Join(Player { name, uuid }))).await?;
					return self.play(tx, rx).await
				}
				ReadPacketOutcome::Eof => {
					debug!("#{} got eof, closing", self.id);
					return Ok(())
				},
				_ => (),
			},
		} }
	}

	async fn play(mut self, tx: Sender, mut rx: Receiver) -> anyhow::Result<()> {
		debug!("#{} entering play", self.id);
		self.write_packet(ServerPacket::JoinGame {
			eid: self.id,
			gamemode: 1,
			hardcode: false,
		}).await?;

		self.write_packet(ServerPacket::PluginMessage {
			channel: "minecraft:brand".to_owned(),
			data: b"\x0bquectocraft".to_vec(),
		}).await?;

		// fabric clients need a 3x3 of chunks to load
		for x in -1..=1 {
			for z in -1..=1 {
				self.write_packet(ServerPacket::ChunkData { x, z }).await?;
			}
		}

		self.write_packet(ServerPacket::SetDefaultSpawn {
			pos: Position { x: 0, y: 0, z: 0 },
			angle: 0.0
		}).await?;

		self.write_packet(ServerPacket::PositionAndLook {
			pos: [0.0, 64.0, 0.0],
			look: [0.0, 0.0],
			flags: 0,
		}).await?;

		loop { select! {
			ev = rx.recv() => match ev {
				Some(ServerEvent::Disconnect(msg)) => {
					self.write_packet(ServerPacket::PlayDisconnect(json!(
						{
							"text": msg
						}
					))).await?;
					info!("#{} disconnecting: {}", self.id, msg.unwrap_or_default());
					return Ok(())
				},
				Some(ServerEvent::SystemMessage(msg, overlay)) => {
					debug!("#{} sending system message: {msg}", self.id);
					self.write_packet(ServerPacket::SystemMessage(msg, overlay)).await?;
				},
				Some(ServerEvent::ChatMessage(msg, name)) => {
					debug!("#{} sending chat message: {msg}", self.id);
					self.write_packet(ServerPacket::ChatMessage(msg, name)).await?;
				},
				Some(ServerEvent::KeepAlive(data)) => {
					debug!("#{} sending keepalive: {data}", self.id);
					self.write_packet(ServerPacket::KeepAlive(data)).await?;
				},
				None => (),
			},
			pk = self.read_packet() => match pk? {
				ReadPacketOutcome::Packet(pk) => match pk {
					ClientPacket::KeepAlive(data) => {
						debug!("#{} recieved keepalive: {data}", self.id);
						tx.send((self.id, ClientEvent::KeepAlive(data))).await?;
					},
					ClientPacket::ChatMessage(msg) => {
						debug!("#{} recieved message: {msg}", self.id);
						tx.send((self.id, ClientEvent::Message(msg))).await?;
					},
					ClientPacket::Command(cmd) => {
						debug!("#{} recieved command: {cmd}", self.id);
						tx.send((self.id, ClientEvent::Command(cmd))).await?;
					},
					_ => (),
				}
				ReadPacketOutcome::None => (),
				ReadPacketOutcome::Eof => return Ok(()),
			},
		} }
	}
}

