use std::sync::Arc;

use log::warn;
use mlua::{Lua, Function, IntoLua, FromLua, Value, Table, AnyUserData};
use tokio::sync::mpsc::{Sender, Receiver};
use uuid::Uuid;

use self::{event::{PluginEvent, ServerEvent}, server::{Server, WrappedServer}, plugin::load_plugins};

pub mod event;
mod plugin;
mod server;

#[derive(Clone, Copy, Eq, PartialEq)]
struct UuidUD(Uuid);

impl<'lua> FromLua<'lua> for UuidUD {
    fn from_lua(value: mlua::Value<'lua>, _lua: &'lua Lua) -> mlua::Result<Self> {
		match value {
			Value::UserData(ud) => Ok(*ud.borrow::<Self>()?),
			v => Err(mlua::Error::FromLuaConversionError { from: v.type_name(), to: "Uuid", message: Some("Expected Uuid".to_owned()) })
		}
    }
}

impl mlua::UserData for UuidUD {
	fn add_methods<'lua, M: mlua::UserDataMethods<'lua, Self>>(methods: &mut M) {
		methods.add_meta_method("__tostring",
			|lua, this, ()| this.0.to_string().into_lua(lua)
		);
		methods.add_meta_method("__eq",
			|lua, this, UuidUD(uuid)| (this.0 == uuid).into_lua(lua)
		);
	}
}

fn broadcast_event<'lua, F>(lua: &'lua Lua, fname: &str, callback: F) -> mlua::Result<()>
where F: Fn(&str, Table<'lua>, AnyUserData, Function) -> mlua::Result<()> {
	let srv: AnyUserData = lua.globals().get("srv")?;
	let plugins: Table = srv.nth_user_value(1)?;
    for (id, table) in plugins.pairs::<String, Table>().filter_map(Result::ok) {
        if let Ok(f) = table.get::<_, Function>(fname) {
			srv.set_nth_user_value(2, id.clone())?;
			if let Err(e) = callback(&id, table.clone(), srv.clone(), f) {
                warn!("error in {fname} in plugin {}: {e}", id);
			}
        }
    }
	Ok(())
}

pub async fn run_plugins(tx: Sender<PluginEvent>, mut rx: Receiver<ServerEvent>) -> mlua::Result<()> {
	let lua = unsafe {
		mlua::Lua::unsafe_new_with(
			mlua::StdLib::ALL_SAFE | mlua::StdLib::DEBUG,
			mlua::LuaOptions::default()
		)
	};

	let server = WrappedServer(Arc::new(Server {
		tx,
		names: Default::default(),
		info: Default::default(),
	}));

	{
		let server_any: AnyUserData = FromLua::from_lua(IntoLua::into_lua(server.clone(), &lua)?, &lua)?;
		let plugins = load_plugins(&lua)?;

		server_any.set_nth_user_value(1, plugins)?;
		server_any.set_nth_user_value(2, Value::Nil)?;
		lua.globals().set("srv", server_any.clone())?;
	}

	broadcast_event(&lua, "init", |_, pl, _, f| {
		f.call(pl)
	})?;

	while let Some(ev) = rx.recv().await {
		match ev {
			ServerEvent::Join(player, info) => {
				server.names.write().await.insert(player.uuid, player.name.clone());
				server.info.write().await.insert(player.uuid, info);
				broadcast_event(&lua, "on_join", |_, pl, _, f| {
					f.call::<_, ()>((
                        pl,
                        UuidUD(player.uuid),
                        player.name.clone()
                    ))
                })?;
			},
			ServerEvent::Leave(player) => {
				broadcast_event(&lua, "on_leave", |_, pl, _, f| {
					f.call::<_, ()>((
                        pl,
                        UuidUD(player.uuid),
                        player.name.clone()
                    ))
                })?;
				server.names.write().await.remove(&player.uuid);
				server.info.write().await.remove(&player.uuid);
			},
			ServerEvent::Message(player, msg) => {
				broadcast_event(&lua, "on_message", |_, pl, _, f| {
					f.call::<_, ()>((
                        pl,
						msg.clone(),
                        UuidUD(player.uuid),
                        player.name.clone()
                    ))
                })?;
			},
			ServerEvent::Command(_, _) => () // TODO command registerment
		}
	}
	Ok(())
}
