use std::{collections::HashMap, sync::Arc};

use log::{Record, warn};
use mlua::{Value, IntoLua, Lua, Table, Function, AnyUserData, LuaSerdeExt};
use serde_json::json;
use tokio::sync::{mpsc::Sender, RwLock};
use uuid::Uuid;

use crate::{ClientInfo, QC_VERSION, JsonValue};

use super::{event::PluginEvent, UuidUD};

impl<'lua> IntoLua<'lua> for ClientInfo {
    fn into_lua(self, lua: &'lua mlua::Lua) -> mlua::Result<Value<'lua>> {
        let table = lua.create_table()?;
		table.set("addr", self.addr)?;
		table.set("port", self.port)?;
		table.set("proto_name", self.proto_name)?;
		table.set("proto_version", self.proto_version)?;
		Ok(Value::Table(table))
    }
}

fn lua_locinfo(lua: &Lua) -> Option<(u32, String)> {
	let debug = lua.globals().get::<_, Table>("debug").ok()?;
	let getinfo = debug.get::<_, Function>("getinfo").ok()?;
	let data = getinfo.call::<_, Table>((2,)).ok()?;
	let line = data.get("currentline").ok()?;
	let file = data.get("short_src").ok()?;
	Some((line, file))
}

fn lua_log(level: log::Level, path: &str, msg: String, lua: &Lua) {
	let mut record = Record::builder();
	record.level(level);
	record.target(path);
	record.module_path(Some(path));
	let locinfo = lua_locinfo(lua);
	if let Some((line, file)) = &locinfo {
		record.line(Some(*line));
		record.file(Some(file));
	}
	log::logger().log(&record.args(format_args!("{}", msg)).build());

}

const LEVELS: [(&str, log::Level); 5] = [
	("trace", log::Level::Trace),
	("debug", log::Level::Debug),
	("info", log::Level::Info),
	("warn", log::Level::Warn),
	("error", log::Level::Error),
];

fn value_to_chat<'lua>(lua: &'lua Lua, value: Value<'lua>) -> mlua::Result<JsonValue> {
	match value {
		Value::String(s) => Ok(json!{{"text": s.to_str()?}}),
		Value::Nil => Ok(json!{{"text":""}}),
		Value::Table(_) => Ok(lua.from_value(value)?),
		_ => return Err(mlua::Error::RuntimeError("()".to_owned())),
	}
}


pub struct Server {
	pub tx: Sender<PluginEvent>,
	pub names: RwLock<HashMap<Uuid, String>>,
	pub info: RwLock<HashMap<Uuid, ClientInfo>>,
}

#[derive(Clone)]
pub struct WrappedServer(pub Arc<Server>);

impl std::ops::Deref for WrappedServer {
	type Target = Server;

	fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl mlua::UserData for WrappedServer {
	fn add_methods<'lua, M: mlua::UserDataMethods<'lua, Self>>(methods: &mut M) {
		// commands
		methods.add_async_method("add_command", |_lua, _this, ()| async {
			warn!("TODO add command");
			Ok(())
		});

		// players
		methods.add_async_method("players", |lua, this, ()| async {
			let table = lua.create_table()?;
			for (k, v) in this.names.read().await.iter() {
				table.set(UuidUD(*k), v.clone())?;
			}
			Ok(table)
		});
		methods.add_async_method("get_name", |lua, this, UuidUD(uuid)| async move {
			match this.info.read().await.get(&uuid) {
				Some(s) => s.clone().into_lua(lua),
				None => Ok(Value::Nil),
			}
		});
		methods.add_async_method("get_info", |lua, this, UuidUD(uuid)| async move {
			match this.info.read().await.get(&uuid) {
				Some(i) => i.clone().into_lua(lua),
				None => Ok(Value::Nil)
			}
		});

		// events
		methods.add_async_method("kick", |_, this, (UuidUD(uuid), msg)| async move {
			this.tx
				.send(PluginEvent::Kick(uuid, msg))
				.await
				.map_err(|e| mlua::Error::RuntimeError(e.to_string()))
		});
		methods.add_async_method("send_msg", |lua, this, (UuidUD(uuid), msg, overlay): (_, Value, Option<bool>)| async move {
			let msg = match msg {
				Value::String(s) => json!{{"text": s.to_str()?}},
				Value::Table(_) => lua.from_value(msg)?,
				_ => return Err(mlua::Error::RuntimeError("()".to_owned())),
			};
			this.tx
				.send(PluginEvent::System(uuid, msg, overlay == Some(true)))
				.await
				.map_err(|e| mlua::Error::RuntimeError(e.to_string()))
		});
		methods.add_async_method("broadcast_msg", |lua, this, (msg, overlay): (Value, Option<bool>)| async move {
			let msg = match msg {
				Value::String(s) => json!{{"text": s.to_str()?}},
				Value::Table(_) => lua.from_value(msg)?,
				_ => return Err(mlua::Error::RuntimeError("()".to_owned())),
			};
			this.tx
				.send(PluginEvent::BroadcastSystem(msg, overlay == Some(true)))
				.await
				.map_err(|e| mlua::Error::RuntimeError(e.to_string()))
		});
		methods.add_async_method("send_chat", |lua, this, (UuidUD(uuid), msg, name): (_, Value, Value)| async move {
			let msg = value_to_chat(lua, msg)?;
			let name = value_to_chat(lua, name)?;
			this.tx
				.send(PluginEvent::Chat(uuid, msg, name))
				.await
				.map_err(|e| mlua::Error::RuntimeError(e.to_string()))
		});
		methods.add_async_method("broadcast_chat", |lua, this, (msg, name): (Value, Value)| async move {
			let msg = value_to_chat(lua, msg)?;
			let name = value_to_chat(lua, name)?;
			this.tx
				.send(PluginEvent::BroadcastChat(msg, name))
				.await
				.map_err(|e| mlua::Error::RuntimeError(e.to_string()))
		});

		// utility
		methods.add_function("check_semver", |_, (ver, req): (String, String)| {
			let ver = semver::Version::parse(&ver)
				.map_err(|_| mlua::Error::runtime(format!("invalid version: {ver}")))?;
			let req = semver::VersionReq::parse(&req)
				.map_err(|_| mlua::Error::runtime(format!("invalid version request: {req}")))?;
			Ok(req.matches(&ver))
		});

		// logging
		for (name, level) in LEVELS {
			methods.add_function(name, move |lua, (this, msg): (AnyUserData, String)| {
				let mut id: String = this.nth_user_value(2)?;
				id.insert_str(0, "plugin::");
				lua_log(level, &id, msg, lua);
				Ok(())
			});
		}
	}

	fn add_fields<'lua, F: mlua::UserDataFields<'lua, Self>>(fields: &mut F) {
		fields.add_field("version", QC_VERSION);
		fields.add_field_function_get("plugins", |_, this| {
			this.nth_user_value::<Table>(1)
		})
	}
}

