use std::io::{Write, Read};

use anyhow::anyhow;
use log::warn;

use super::{ServerPacket, ClientPacket};
use crate::{varint::VarInt, ser::{Serializable, Deserializable}};

use super::{Protocol, ProtocolState};

pub const COMMON: Protocol = Protocol {
	name: "common",
	version: i32::MAX,
	encode,
	decode,
};

fn encode(mut w: Box<&mut dyn Write>, _state: ProtocolState, pk: ServerPacket) -> anyhow::Result<()> {
	match pk {
		ServerPacket::StatusResponse(response) => {
			VarInt(0x00).serialize(&mut w)?;
			response.serialize(&mut w)?;
		},
		ServerPacket::PingResponse(payload) => {
			VarInt(0x01).serialize(&mut w)?;
			payload.serialize(&mut w)?;
		},
		ServerPacket::LoginDisconnect(response) => {
			VarInt(0x00).serialize(&mut w)?;
			response.serialize(&mut w)?;
		},
		_ => {
			return Err(anyhow!("packet {pk:?} unhandled"))
		}
	}
	Ok(())
}

fn decode(mut r: Box<&mut dyn Read>, state: ProtocolState, _len: i32, id: i32) -> anyhow::Result<Option<ClientPacket>> {
	type Ps = ProtocolState;
	match (state, id) {
		(Ps::Handshake, 0x00) => {
			let version = VarInt::deserialize(&mut r)?.0;
			let addr = String::deserialize(&mut r)?;
			let port = u16::deserialize(&mut r)?;
			let next_state = VarInt::deserialize(&mut r)?.0;
			Ok(Some(ClientPacket::Handshake { version, addr, port, next_state }))
		},
		(Ps::Status, 0x00) => {
			Ok(Some(ClientPacket::StatusRequest))
		},
		(Ps::Status, 0x01) => {
			let payload = i64::deserialize(&mut r)?;
			Ok(Some(ClientPacket::PingRequest(payload)))
		},
		_ => {
			warn!("invalid id {id:#x} in state {state:?}");
			Ok(None)
		}
	}
}
