use std::{collections::HashMap, time::Duration, net::{ToSocketAddrs, SocketAddr}};

use log::{info, error};
use tokio::{net::{TcpListener, TcpStream}, sync::mpsc::{Sender, self}, select};
use uuid::Uuid;
use anyhow::anyhow;

use crate::{client::run_client, Player, ClientInfo, client::event::{ServerEvent, ClientEvent}, plugin::{run_plugins, event::PluginEvent}};

type PServerEvent = crate::plugin::event::ServerEvent;

pub struct Client {
	tx: Sender<ServerEvent>,
	keepalive: Option<i64>,
	info: Option<ClientInfo>,
	player: Option<Player>,
}

struct Server {
	gc_tx: Sender<(i32, ClientEvent)>,
	ps_tx: Sender<PServerEvent>,
	next_id: i32,
	clients: HashMap<i32, Client>,
	ids: HashMap<Uuid, i32>,
}

pub async fn run_server() -> anyhow::Result<()> {
	env_logger::init();

	let (ps_tx, ps_rx) = mpsc::channel(256);
	let (pc_tx, mut pc_rx) = mpsc::channel(256);

	tokio::spawn(async {
		if let Err(e) = run_plugins(pc_tx, ps_rx).await {
			println!("error in plugins: {e}");
		}
	});

	let socket_addr = "0.0.0.0:25567".to_socket_addrs()?
		.next().expect("invalid server address");
	let listener = TcpListener::bind(socket_addr).await?;

	let (gc_tx, mut gc_rx) = mpsc::channel(16);

	let mut server = Server {
		gc_tx, ps_tx,
		next_id: 0,
		clients: HashMap::new(),
		ids: HashMap::new(),
	};

	let mut keepalive = tokio::time::interval(Duration::from_secs(15));

	info!("listening on {socket_addr}");

	loop { select! {
		conn = listener.accept() => {
			let (stream, addr) = conn?;
			accept_connection(&mut server, stream, addr).await?;
		},
		_ = keepalive.tick() => handle_keepalive(&mut server).await?,
		ev = pc_rx.recv() => {
			let Some(ev) = ev else {
				return Err(anyhow!("plugin channel closed"));
			};
			handle_plugin_event(&mut server, ev).await?;
		},
		ev = gc_rx.recv() => {
			let Some(ev) = ev else {
				return Err(anyhow!("reciever closed"));
			};
			handle_client_event(&mut server, ev.0, ev.1).await?;
		}
	} }
}

async fn accept_connection(server: &mut Server, stream: TcpStream, addr: SocketAddr) -> anyhow::Result<()> {
	let gc_tx = server.gc_tx.clone();

	let id = server.next_id;
	server.next_id = server.next_id.wrapping_add(1);

	let (gs_tx, gs_rx) = mpsc::channel(16);
	server.clients.insert(id, Client {
		tx: gs_tx,
		keepalive: None,
		info: None,
		player: None,
	});
	info!("#{id} connected from {addr}");
	tokio::spawn(async move {
		let c_tx2 = gc_tx.clone();
		match run_client(id, stream, gc_tx, gs_rx).await {
			Ok(_) => info!("#{id} disconnected"),
			Err(e) => error!("#{id} disconnected: {e}"),
		}
		c_tx2.send((id, ClientEvent::Disconnect)).await.unwrap();
	});
	Ok(())
}

async fn handle_keepalive(server: &mut Server) -> anyhow::Result<()> {
	for client in server.clients.values_mut() {
		if client.keepalive.is_some() {
			client.tx.send(ServerEvent::Disconnect(Some("Failed to respond to keep alives".to_owned()))).await?;
		} else if client.player.is_some() {
			let data = 1;
			client.keepalive = Some(data);
			client.tx.send(ServerEvent::KeepAlive(data)).await?;
		}
	}
	Ok(())
}

async fn handle_plugin_event(server: &mut Server, ev: PluginEvent) -> anyhow::Result<()> {
	match ev {
		PluginEvent::Kick(uuid, msg) => {
			if let Some(id) = server.ids.get(&uuid) {
				server.clients[id].tx
					.send(ServerEvent::Disconnect(msg))
					.await?;
			}
		},
		PluginEvent::Chat(uuid, msg, name) => {
			if let Some(id) = server.ids.get(&uuid) {
				server.clients[id].tx
					.send(ServerEvent::ChatMessage(msg, name))
					.await?;
			}
		},
		PluginEvent::BroadcastChat(msg, name) => {
			for client in server.clients.values() {
				if client.player.is_some() {
					client.tx.send(ServerEvent::ChatMessage(msg.clone(), name.clone())).await?;
				}
			}
		},
		PluginEvent::System(uuid, msg, overlay) => {
			if let Some(id) = server.ids.get(&uuid) {
				server.clients[id].tx
					.send(ServerEvent::SystemMessage(msg, overlay))
					.await?;
			}
		},
		PluginEvent::BroadcastSystem(msg, overlay) => {
			for client in server.clients.values() {
				if client.player.is_some() {
					client.tx.send(ServerEvent::SystemMessage(msg.clone(), overlay)).await?;
				}
			}
		},
	}
	Ok(())
}

async fn handle_client_event(server: &mut Server, id: i32, ev: ClientEvent) -> anyhow::Result<()> {
	match ev {
		ClientEvent::Handshake(info) => {
			server.clients.get_mut(&id).unwrap().info = Some(info)
		},
		ClientEvent::Join(player) => {
			if let Some(other_id) = server.ids.remove(&player.uuid) {
				let other_client = server.clients.remove(&other_id).unwrap();
				server.ps_tx
					.send(PServerEvent::Leave(other_client.player.unwrap()))
					.await?;
				other_client.tx
					.send(ServerEvent::Disconnect(Some("logged in elsewhere".to_owned())))
					.await?;
			}
			let client = server.clients.get_mut(&id).unwrap();
			client.player = Some(player.clone());
			server.ids.insert(player.uuid, id);
			server.ps_tx.send(PServerEvent::Join(player, client.info.clone().unwrap())).await?;
		},
		ClientEvent::Disconnect => if let Some(client) = server.clients.remove(&id) {
			if let Some(player) = client.player {
				server.ids.remove(&player.uuid);
				server.ps_tx.send(PServerEvent::Leave(player)).await?;
			}
		},
		ClientEvent::KeepAlive(data) => {
			let client = server.clients.get_mut(&id).unwrap();
			if client.keepalive == Some(data) {
				client.keepalive = None;
			} else {
				client.keepalive = Some(0);
			}
		},
		ClientEvent::Command(cmd) => {
			let player = server.clients.get(&id).unwrap().player.clone().unwrap();
			server.ps_tx.send(PServerEvent::Command(player, cmd)).await?;
		},
		ClientEvent::Message(msg) => {
			let player = server.clients.get(&id).unwrap().player.clone().unwrap();
			server.ps_tx.send(PServerEvent::Message(player, msg)).await?;
		},
	}
	Ok(())
}
